from server import server
import asyncio
import socket


if __name__ == "__main__":
    #try:
    server_obj = server.Server('127.0.0.1', 30000)
    asyncio.run(server_obj.main())
    # except KeyboardInterrupt:
    #     if server_obj:
    #         server_obj.running = False



def cast_int(data):
    return int(data)

def cast_string(data):
    return str(data)

def cast_bool(data):
    return bool(data)



class Validators:
    type_map = {
        'int':cast_int,
        'string':cast_string,
        'bool':cast_bool
    }
    def __init__(self, arg_type):
        self.validator_map = {
            'silent':self.no_validator,
            'secure':self.standard_validator,
            'expression':self.standard_validator,
            'str_input':self.standard_validator,
            'standard':self.standard_validator,
            'confirmation':self.confirmation_validator,
            'input_list':self.list_validator,
            'input_dict':self.dict_validator,
            'form':self.form_validator,
            'selection_list':self.selection_list_validator
        }

    def set_default_if_needed(self, value):
        """
        if there is no value, set the default value. If the value is False, return False
        """
        if value == None and self.default_value:
            value = self.default_value
            return (True, value)
        elif not value:
            return (True, value)
        return (False, value)
    
    def check_data_type_followed(self, value):
        """
        check if the value is of the correct datatype
        """
        try:
            return self.type_map[self.data_type](value)
        except:
            raise ValueError(f"The argument {self.argument} requires a datatype {self.data_type}")
        
    def check_size_limit_followed(self, value):
        """
        check that the value abids byt he size limit
        """
        min_, max_ = self.size_limit
        if min_ != -1 and max_ != -1:
            if type(value) == int:
                if value >= max_ or value < min_:
                    raise ValueError(f"The argument {self.argument} must be a value in the range {self.size_limit}")
            elif type(value) == str and (len(value) >= max_ or len(value) < min_):
                raise ValueError(f"The argument {self.argument} must be between the sizes {self.size_limit}")

    def no_validator(self, value):
        return value
    
    def standard_validator(self, value):
        """
        check to make sure the value assigned to the argument follows all existing rules
        """
        return_default, value = self.set_default_if_needed(value)
        if return_default:
            return value
        value = self.check_data_type_followed(value)
        self.check_size_limit_followed(value)
        if self.choices and value not in self.choices:
            raise ValueError(f"The value for argument {self.argument} must be in the list {self.choices}")
        if self.arg_type != 'bool' and isinstance(value, bool):
            return None
        return value
    
    def confirmation_validator(self, value):
        if self.required and not value:
            raise RuntimeError("Declined confirmation, cancelling")
        return value
    
    def list_validator(self, value):
        if value and not isinstance(value, bool):
            processed_values = list()
            for sub_value in value:
                processed_values.append(self.data_type.verify_rules_followed(sub_value))
            return processed_values
        if isinstance(value, bool) and value:
            return True
        else:
            return None
    
    def dict_validator(self, value):
        if value and not isinstance(value, bool):
            processed_values = dict()
            for sub_name, sub_value in value.items():
                processed_values[sub_name] = self.data_type.verify_rules_followed(sub_value)
            return processed_values
        if isinstance(value, bool) and value:
            return True
        else:
            return None
    
    def form_validator(self, value):
        if value and not isinstance(value, bool):
            flattened_value = self.flatten_form_data(value)
            results = dict()
            for argument_name, argument in self.arguments_list.items():
                results[argument_name] = argument.verify_rules_followed(flattened_value[argument_name])
            return flattened_value
        if isinstance(value, bool) and value:
            return True
        else:
            return None
    
    def selection_list_validator(self, value):
        if value and not isinstance(value, bool):
            results = dict()
            for field, field_bool in value.items():
                if field_bool:
                    results[field] = self.option_dict[field]
        else:
            return None
        return results

        



from .. import baseCommand, decorators
from ..arguments import argument
Argument = argument.Argument


CONFIGURE_RUNTIME = argument.Form('configureRuntime', flattening_type='RETRIEVE', required_arguments = [
        Argument('runtime', choices=['SINGULARITY', 'DOCKER']),
        Argument('runtimeVersion')
    ],
    optional_arguments=[
        Argument('runtimeOptions', arg_type='input_list', data_type=Argument('runtimeOption',
            arg_type='str_input', choices=['NONE', 'SINGULARITY_START', 'SINGULARITY_RUN'])),
])

JOB_CONSTRAINTS = argument.Form('configureJobSettings', flattening_type='RETRIEVE', required_arguments = [
    Argument('jobType', choices=['BATCH', 'FORK']),
    Argument('maxJobs', data_type='int'),
    Argument('maxJobsPerUser', data_type='int'),
])

SYSTEM_CONFIG = argument.Form('defaultSystemConfig', flattening_type='RETRIEVE', optional_arguments = [
    Argument('dynamicExecSystem', arg_type='confirmation', description="System is dynamic?"),
    Argument('execSystemConstraints', arg_type='input_list', data_type=Argument('constraint', size_limit=(3, 4096))),
    Argument('execSystemExecDir', size_limit=(1, 4096), description='based on the system job working dir? automatically generated if not specified'),
    Argument('execSystemInputDir', size_limit=(1, 4096), description='what system path will be used to stage input files? automatically generated if not specified'),
    Argument('execSystemOutputDir', size_limit=(1, 4096), description='Where will tapis put job output files? automatically generated if not specified'),
    Argument('execSystemLogicalQueue', size_limit=(1, 128), description='What batch logical queue on the system will be used for execution?'),
])

ARCHIVE_ON_APP_ERROR = argument.Form('archiveOnAppError', flattening_type='FLATTEN',  required_arguments=[
    Argument('archiveSystemId', size_limit=(1, 80), description='What system will be used when archiving outputs?'),
    Argument('archivesystemDir', size_limit=(1, 4096), description='What system directory will be used for archiving?'),
])


APP_ARGUMENTS = Argument('appArgs', arg_type='input_list', data_type=argument.Form('appArg', required_arguments=[
            Argument('name', size_limit=(1, 80)),
            Argument('inputMode', choices=['REQUIRED', "FIXED", "INCLUDE_ON_DEMAND", "INCLUDE_BY_DEFAULT"]),
            Argument('arg')
        ],
        optional_arguments=[
            Argument('description', size_limit=(1, 8096))
        ]
    ), description='command line arguments to be passed to the application', part_of='parameterSet')


CONTAINER_ARGUMENTS = Argument('containerArgs', arg_type='input_list', data_type=argument.Form('containerArg', required_arguments=[
            Argument('name', size_limit=(1, 80)),
            Argument('arg', description='value for the argument'),
            Argument('inputMode', choices=['REQUIRED', "FIXED", "INCLUDE_ON_DEMAND", "INCLUDE_BY_DEFAULT"], description='How will this argument be treated when processing jobs? EG should the job require it?')
        ],
        optional_arguments=[
            Argument('description', size_limit=(1, 8096))
        ]
    ), description='command line arguments to be passed to the container at runtime, whether singularity or container', part_of='parameterSet')


SCHEDULER_OPTIONS = Argument('schedulerOptions', arg_type='input_list', data_type=argument.Form('schedulerOption', required_arguments=[
            Argument('name', size_limit=(1, 80)),
            Argument('inputMode', choices=['REQUIRED', "FIXED", "INCLUDE_ON_DEMAND", "INCLUDE_BY_DEFAULT"]),
            Argument('arg')
        ],
        optional_arguments=[
            Argument('description', size_limit=(1, 8096)),
        ]
    ), description='options to pass to hpc batch scheduler', part_of='parameterSet')


JOB_ENVIRONMENT_VARIABLES = Argument('envVariables', arg_type='input_list', data_type=argument.Form('environment_variable', required_arguments=[
            Argument('key'),
            Argument('value'),
        ],
        optional_arguments=[
            Argument('description', size_limit=(1, 2048))
        ]
    ), description='environment variables placed into the runtime environment', part_of='parameterSet')

JOB_ALLOCATION_CONFIGURATION = argument.Form('jobAllocationConfiguration', flattening_type='RETRIEVE', required_arguments = [
    Argument('nodeCount', data_type='int', description='how many nodes do you want to request in your job?', default_value=1),
    Argument('coresPerNode', data_type='int', description='how many cores per node?', default_value=1),
    Argument('memoryMB', data_type='int', default_value=1),
    Argument('maxMinutes', data_type='int', description='max job runtime in minutes', part_of='parameterSet', default_value=10),
])


import typing
import abc
import pprint

from . import argumentValidators


ALLOWED_ARG_TYPES = typing.Literal['silent', 'secure', 'expression', 'input_list', 'input_dict', 'form', 'str_input', 'confirmation', 'selection_list']
ALLOWED_DATA_TYPES = typing.Literal['string', 'int', 'bool']
ALLOWED_ACTIONS = typing.Literal['store', 'store_true', 'store_false']


class AbstractArgument(abc.ABC):
    @abc.abstractmethod
    def json(self):
        pass
    

class Argument(argumentValidators.Validators, AbstractArgument):
    def __init__(self, argument: str,
                 data_type: ALLOWED_DATA_TYPES | typing.Type[AbstractArgument] = 'string',
                 arg_type: ALLOWED_ARG_TYPES | typing.Literal['standard']='standard',
                 choices: list | None=None, 
                 action: typing.Literal['store', 'store_true', 'store_false']="store", 
                 description: str="",
                 positional: bool=False,
                 default_value=None,
                 depends_on: list = [],
                 mutually_exclusive_with: list = [],
                 part_of: str = "",
                 size_limit: tuple=(-1,-1)):
        if arg_type not in typing.get_args(ALLOWED_ARG_TYPES) and arg_type != 'standard':
            raise ValueError(f"The arg type parameter in the argument {self.__class__.__name__} must be in the list {ALLOWED_ARG_TYPES}. Got arg type {arg_type}")
        if data_type not in typing.get_args(ALLOWED_DATA_TYPES) and not issubclass(data_type.__class__, AbstractArgument):
            raise ValueError(f"The data type argument provided to the argument {self.__class__.__name__} must be in the list {ALLOWED_DATA_TYPES}, or must be a subclass of AbstractArgument")
        if action not in typing.get_args(ALLOWED_ACTIONS):
            raise ValueError(f"Action {action} not in the list {ALLOWED_ACTIONS}, in argument {self.__class__.__name__}")
        if arg_type in ('input_list', 'input_dict') and isinstance(data_type, Argument) and data_type.arg_type == 'standard':
            data_type.arg_type = 'str_input'
        if action != 'store':
            data_type = 'bool'
        super().__init__(arg_type)
        if action in ('store_true', 'store_false'):
            data_type = 'bool'
        if arg_type != 'standard':
            action = 'store_true'
        self.argument = argument
        self.arg_type = arg_type
        self.data_type = data_type
        self.choices = choices
        self.action = action
        self.description = description
        self.positional = positional
        self.default_value = default_value
        self.size_limit=size_limit
        self.depends_on = depends_on
        self.mutually_exclusive_with = mutually_exclusive_with
        self.required = None
        self.part_of = part_of

        self.truncated_arg = None
        self.full_arg = f"--{self.argument}"

    def is_required(self, is_required: bool):
        """
        setter for the required attribute
        """
        self.required = is_required

    def verify_rules_followed(self, value):
        """
        check to make sure the value assigned to the argument follows all existing rules
        """
        if self.required and not value and value != False and self.arg_type != 'silent':
            raise ValueError(f'The argument {self.argument} is required')
        return_value = self.validator_map[self.arg_type](value)
        if isinstance(return_value, bool) and self.data_type != 'bool' and self.action != 'store_true' and not return_value:
            print(f"{self.argument} RETURNING NONE")
            return None
        return return_value

    def json(self):
        json = {
            'name':self.argument,
            'arg_type':self.arg_type,
            'data_type':self.data_type,
            'choices':self.choices,
            'action':self.action,
            'description':self.description,
            'positional':self.positional,
            'default_value':self.default_value,
            'size_limit':self.size_limit,
            'truncated_arg':self.truncated_arg,
            'full_arg':self.full_arg,
            'depends_on':self.depends_on,
            'mutually_exclusive_with':self.mutually_exclusive_with,
            'part_of':self.part_of,
            'required':self.required
        }
        if self.data_type in ('string', 'int', 'bool'):
            json['data_type'] = self.data_type
        else:
            json['data_type'] = self.data_type.json() 
        return json
    
    def help_message(self):
        help = {"name":self.argument,
                "description":f"{self.description}"}
        if self.choices and isinstance(self.choices, list):
            help['choices'] = self.choices
        if self.positional:
            help['syntax'] = f"<{self.argument}>"
        elif self.action == 'store':
            help['syntax'] = f"{self.truncated_arg}/{self.full_arg} <{self.argument}>"
        else:
            help['syntax'] = f"{self.truncated_arg}/{self.full_arg}"
            if self.arg_type != 'standard':
                help['syntax'] = help['syntax'] + " (f)"
        return help
    
    def check_for_copy_data(self):
        return {
            'name':self.argument,
            'action':self.action,
        }
    
    def str(self):
        help_str = f"{self.truncated_arg}/{self.full_arg} "
        if self.positional:
            help_str = f"<{self.argument}> "
        elif self.action == 'store':
            help_str += f"<{self.argument}> "
        if self.arg_type != 'standard':
            help_str += ' (f) '
        return help_str


class Form(Argument):
    def __init__(self, form_name, required_arguments=None, optional_arguments=None, description=None, depends_on=None, flattening_type: typing.Literal['NONE', 'FLATTEN', 'RETRIEVE']='NONE'):
        super().__init__(form_name, arg_type='form', description=description, depends_on=depends_on)
        arguments_list = list()
        if required_arguments:
            for argument in required_arguments:
                argument.is_required(True)
            arguments_list += required_arguments
        if optional_arguments:
            for argument in optional_arguments:
                argument.is_required(False)
            arguments_list += optional_arguments
        for argument in arguments_list:
            if not issubclass(argument.__class__, Argument):
                raise TypeError(f'The argument {argument.argument} must be an Argument type')
            if argument.arg_type == 'standard':
                argument.arg_type = 'str_input'
        self.arguments_list = {argument.argument:argument for argument in arguments_list}
        self.flattening_type = flattening_type
    
    def flatten_form_data(self, value): 
        if value and not isinstance(value, bool):
            if self.flattening_type == 'FLATTEN':
                flattened_form_info = {self.argument:True}
                flattened_form_info.update(**{arg_name:arg_value for arg_name, arg_value in value[self.argument].items()})
                return flattened_form_info
            if self.flattening_type == 'RETRIEVE':
                exctracted_form_data = {arg_name:arg for arg_name, arg in value[self.argument].items()}
                return exctracted_form_data
            else:
                return value[self.argument]

    def json(self):
        fields = {argument_name:argument.json() for argument_name, argument in self.arguments_list.items()}
        json = {
            'name':self.argument,
            'arg_type':self.arg_type,
            'data_type':self.data_type,
            'choices':self.choices,
            'action':self.action,
            'description':self.description,
            'positional':self.positional,
            'default_value':self.default_value,
            'size_limit':self.size_limit,
            'truncated_arg':self.truncated_arg,
            'full_arg':self.full_arg,
            'arguments_list':fields,
            'required':self.required,
            'flattening_type':self.flattening_type
        }
        return json
    

class SelectionList(Argument):
    def __init__(self, argument, option_dict):
        super().__init__(argument, data_type='selection_list')
        self.option_dict = option_dict
        self.option_list = list(option_dict.keys())

    def json(self):
        json = {
            'name':self.argument,
            'arg_type':self.arg_type,
            'data_type':self.data_type,
            'choices':self.choices,
            'action':self.action,
            'description':self.description,
            'positional':self.positional,
            'default_value':self.default_value,
            'size_limit':self.size_limit,
            'truncated_arg':self.truncated_arg,
            'full_arg':self.full_arg,
            'option_list':self.option_list,
            'required':self.required
        }
        return json
    

class RequestHandler:
    def handle_requests(self, request):
        pass
    

if __name__ == "__main__":
    pass
    




import typing
import webbrowser

from tapipy.tapis import Tapis


if __name__ != "__main__":
    from . import baseCommand, decorators
    from utilities import exceptions
    from commands.arguments.argument import Argument


class ServiceChecker:
    def __init__(self, available_services: list):
        self.available_services = available_services

    def check_services(self, t):
        tenant = t.tenants.get_tenant(tenant_id=t.tenant_id)
        site_id = tenant.site_id
        supported_services = t.tenants.get_site(site_id=site_id).services
        filtered_supported_services = [service for service in supported_services if service in self.available_services]
        return filtered_supported_services


class get_tenants(baseCommand.BaseCommand): # give this return fields
    """
    @help: get a list of available tenants to authenticate with
    """
    async def run(self, *args, **kwargs):
        return_data = dict()
        tenants = self.t.tenants.list_tenants()
        for tenant in tenants:
            return_data[tenant.tenant_id] = {'uri':tenant.base_url.split('//')[1], 'owner':tenant.owner, 'description':tenant.description}
        return return_data

    

class get_tenant(baseCommand.BaseCommand):
    """
    @help: get a specific tenant
    """
    required_arguments = [
        Argument('tenant_id', positional=True)
    ]
    async def run(self, *args, **kwargs):
        tenant = self.t.tenants.get_tenant(**kwargs)
        return {'uri':tenant.base_url.split('//')[1], 'owner':tenant.owner, 'description':tenant.description}
    
    
class switch_tenant_to(baseCommand.BaseCommand):
    """
    @help: switch the connected tapis service
    @todo: upgrade to federated auth
    """
    required_arguments=[
        Argument('tenant_uri', size_limit=(0, 80), positional=True),
        Argument('auth', choices=['password', 'device_code', 'federated']),
    ]
    async def run(self, *args, **kwargs):  # tenant_uri is the baseURL
        auth = kwargs['auth']
        tenant_uri = kwargs['tenant_uri']
        self.server.auth_type = auth
        if auth == "password":
            results = await self.server.password_grant(tenant_uri, kwargs['connection'])
        elif auth == "device_code":
            results = await self.server.device_code_grant(tenant_uri, kwargs['connection'])
        elif auth == "federated":
            results = await self.server.federated_grant(tenant_uri, kwargs['connection'])
        else:
            results = None
        self.server.configure_decorators(self.server.username, self.server.password)
        self.server.update_credentials(self.server.t, self.server.username, self.server.password)
        return results
      

class exit(baseCommand.BaseCommand):
    """
    @help: exit the CLI without shutting down the service
    """
    async def run(self, *args, **kwargs):
        raise exceptions.Exit
    

class shutdown(baseCommand.BaseCommand):
    """
    @help: exit the CLI and shutdown the service
    """
    async def run(self, *args, **kwargs):
        raise exceptions.Shutdown
    
    
class whoami(baseCommand.BaseCommand):
    """
    @help: returns the username of the current user
    """
    async def run(self, *args, **kwargs) -> str:
        user_info = self.t.authenticator.get_userinfo()
        return user_info
    

class user(baseCommand.BaseCommand):
    """
    @help: returns the username of the selected user
    """
    required_arguments = [
        Argument('username', positional=True)
    ]
    async def run(self, *args, **kwargs):
        return self.t.authenticator.get_profile(**kwargs)
    

class whereami(baseCommand.BaseCommand):
    """
    @help: get the URI of current tapis tenant
    """
    async def run(self, *args, **kwargs):
        return self.server.url
    

class manpages(baseCommand.BaseCommand):
    """
    @help: get a link to the application manpages
    """
    async def run(self, *args, **kwargs):
        webbrowser.open('https://github.com/sdsc-hpc-training-org/hello_icicle_auth_clients/tree/main/icicle_rel_03_2023/CLI/TapisCL-ICICLE')
        return "Please see https://github.com/sdsc-hpc-training-org/hello_icicle_auth_clients/tree/main/icicle_rel_03_2023/CLI/TapisCL-ICICLE for manpages (will add actual manpages later)"


import socket
import asyncio
from TapisCLICICLE.client.cli import CLI


def client_start():
    client_object = CLI('127.0.0.1', 30000)
    client_object.main()


if __name__ == "__main__":
    client_start()

# server = Server(socket.gethostbyname(socket.gethostname()), 30000)
# asyncio.run(server.main())


#!/usr/bin/python3
import socket
import argparse
import sys
import os
import time
import traceback
import subprocess
import pprint

import pyfiglet
from blessed import Terminal

if __name__ != "__main__":
    from . import handlers
    from ..socketopts import socketOpts, schemas
    from ..commands import decorators
    from ..utilities import killableThread, exceptions


__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(f"{__file__}")))
server_path = os.path.join(__location__, r'../serverRun.py')

ENTRANCE_MESSAGE = \
"""
Welcome to TapisCLICICLE
The server takes a bit of time to start. Please Wait.
If you have VPN enabled and startup fails, disable your VPN
If issues persist, try restarting your computer.
If you find any issues, please create a new issue here: https://github.com/sdsc-hpc-training-org/hello_icicle_auth_clients/issues
"""
    

class ClientSideConnection(socketOpts.ClientSocketOpts, handlers.Handlers):
    def __init__(self, connection, debug=False):
        super().__init__("client", debug=debug)
        self.connection = connection
    
    def close(self):
        self.connection.close()


class CLI(handlers.Handlers):
    debug=False
    """
    Receive user input, either direct from bash environment or from the custom interface, then parse these commands and send them to the server to be executed. 
    """
    def __init__(self, IP: str, PORT: int):
        super().__init__()

        self.term = Terminal()
        print(f"{self.term.color_rgb(0, 0, 255)}{ENTRANCE_MESSAGE}{self.term.normal}")
        self.ip, self.port = IP, PORT
        self.connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM) 

        # set up argparse
        self.parser = None

        setup_message = schemas.ResponseData(request_content={'setup_success':True})
        try:
            self.username, self.url = self.connect()
            self.connection.send(setup_message)
        except ConnectionAbortedError:
            print("Server timed out during authentication. Try again")
            sys.exit(0)
        except (KeyboardInterrupt, Exception) as e:
            error_str = traceback.format_exc()
            if self.debug:
                print(error_str)
            print(e)
            setup_message.error = str(e)
            setup_message.request_content['setup_success'] = False
            self.connection.send(setup_message)
            sys.exit(0)

        self.pwd = ""
        self.current_system = ""

    def parser_error(self, args):
        print(f"Ignoring unrecognized arguments: {args}")
    
    def configure_parser(self, arguments):
        parser = argparse.ArgumentParser(description="Command Line Argument Parser", exit_on_error=False, usage=argparse.SUPPRESS, add_help=False, conflict_handler='resolve')
        parser.add_argument('command_selection')
        parser.add_argument('positionals', nargs='*')
        parser.error = self.parser_error
        
        for arg_name, arg in arguments.items():
            if (arg['action'] == 'store_true' and arg['arg_type'] == 'standard') or (arg['arg_type'] == 'form' and arg['flattening_type'] == 'FLATTEN'):
                default = False
            elif (arg['action'] == 'store_false' and arg['arg_type'] == 'standard') or (arg['arg_type'] != 'standard' and arg['required']):
                default = True
            else:
                default = None
            parser.add_argument(f"-{arg['truncated_arg']}", arg['full_arg'],
                                action=arg['action'], default=default)
        return parser

    def initialize_server(self): 
        """
        detect client operating system. The local server intitialization is different between unix and windows based systems
        """
        if 'win' in sys.platform:
            os.system(rf"pythonw {server_path} 1>stdout.txt 2>stderr.txt")
        else: # unix based
            subprocess.Popen(['nohup', 'python3', server_path, '&'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    #@decorators.AnimatedLoading
    def connection_initialization(self):
        """
        start the local server through the client
        """
        startup_flag = False
        timeout_time = time.time() + 30 # server setup timeout. If expires, there is a problem!
        while True:
            if time.time() > timeout_time: # connection timeout condition
                sys.stdout.write("\r[-] Connection timeout")
                sys.exit(0)
            try:
                self.connection.connect((self.ip, self.port)) 
                self.connection = ClientSideConnection(self.connection, debug=self.debug)
                if startup_flag:
                    startup.kill()
                break
            except Exception as e:
                if not startup_flag:
                    startup = killableThread.KillableThread(target=self.initialize_server, daemon=True) # run the server setup on a separate thread
                    startup.start() 
                    startup_flag = True # set the flag to true so the thread runs only once
                    continue

    def auth(self):
        while True:
            server_auth_request: schemas.AuthRequest = self.connection.receive()
            if not server_auth_request.message and not server_auth_request.request_content:
                self.print_response(server_auth_request.error)
                sys.exit(0)
            elif server_auth_request.auth_request_type == "success":
                self.print_response(server_auth_request.message)
                return server_auth_request.message['username'], server_auth_request.message['url']
            form_response = self.universal_message_handler(server_auth_request, self.term)
            if not form_response:
                break
            client_response_request = schemas.AuthRequest(auth_request_type=server_auth_request.auth_request_type, request_content=form_response)
            self.connection.send(client_response_request)
        
    def connect(self):
        """
        connect to the local server
        """
        self.connection_initialization() 
        #self.connection.connect((self.ip, self.port)) # enable me for debugging. Requires manual server start
        connection_info: schemas.StartupData = self.connection.receive() # receive info from the server whether it is a first time connection
        if connection_info.initial: # if the server is receiving its first connection for the session\
            username, url = self.auth()
        else:
            username, url = connection_info.username, connection_info.url
        arguments: schemas.ResponseData = self.connection.receive()
        self.parser = self.configure_parser(arguments.request_content)

        return username, url # return the username and url

    def interface(self, kwargs):
        if not kwargs['command_selection']:
            return
        command_request = schemas.CommandData(request_content=kwargs)
        self.connection.send(command_request)
        while True:
            command_response = self.connection.receive()
            if isinstance(command_response, schemas.ResponseData):
                self.url, self.username = command_response.url, command_response.active_username
                self.pwd, self.current_system = command_response.pwd, command_response.system
                if command_response.exit_status:
                    raise exceptions.Shutdown()
            handled_response = self.universal_message_handler(command_response, self.term)
            if not handled_response:
                break
            handled_response = schemas.FormResponse(request_content=handled_response)
            self.connection.send(handled_response)

    def terminal_cli(self):
        try:
            kwargs = self.parser.parse_args()
            kwargs = vars(kwargs)
            self.interface(kwargs)
            kwargs = vars(self.parser.parse_args(['exit']))
            response_message = schemas.CommandData(request_content=kwargs)
            self.connection.send(response_message)
            self.connection.close()
        except Exception as e:
            error_str = traceback.format_exc()
            #if self.debug:
            print(error_str)
            print(e)
            error_message = schemas.ResponseData(error=str(e))
            self.connection.send(error_message)
        finally:
            sys.exit(0)

    def cli_window(self):
        title = pyfiglet.figlet_format("-----------\nTapisCLICICLE\n-----------", font="slant") # print the title when CLI is accessed
        print(title)
        print(r"""Enter 'exit' to exit the client
                Enter 'shutdown' to shut down the client and server
                Enter 'Help' to show command options""")
        while True:
            try:
                time.sleep(0.01)
                kwargs: dict = vars(self.parser.parse_args(str(input(f"[{self.username}@{self.url}][{self.current_system}]{self.pwd} ")).split(" ")))
                self.interface(kwargs)
            except KeyboardInterrupt:
                continue
            except exceptions.Shutdown:
                print("Exit initiated by user")
                kwargs = vars(self.parser.parse_args(['shutdown']))
                response_message = schemas.CommandData(request_content=kwargs)
                self.connection.send(response_message)
                self.connection.close()
                break
            except (ConnectionAbortedError, ConnectionResetError):
                print("Server shutdown, exiting (SF)")
                self.connection.close()
                break
            except (TypeError, argparse.ArgumentError, argparse.ArgumentTypeError) as e:
                print(e)
                error_str = traceback.format_exc()
                print(error_str)
                print("Invalid command")
            except Exception as e:
                error_str = traceback.format_exc()
                #if self.debug:
                print(error_str)
                print(e)
                error_message = schemas.ResponseData(error=str(e))
                self.connection.send(error_message)
        

    def main(self):
        if len(sys.argv) > 1: # checks if any command line arguments were provided. Does not open CLI
            self.terminal_cli()
        self.cli_window()


if __name__ == "__main__":
    client = CLI(socket.gethostbyname(socket.gethostname()), 30000)
    client.main()


import json

import baseCommand
import arguments.argument as argument


class baseConfigGenerator(baseCommand.BaseCommand):
    required_arguments = [
        argument.SelectionList('appServiceOptions', option_dict={
            ''
        })
    ]


from .. import baseCommand, decorators
from ..arguments import argument
from .. import commandOpts
from socketopts import schemas
Argument = argument.Argument


CAN_EXEC = argument.Form('canExec', flattening_type='FLATTEN', required_arguments=[
    Argument('jobRuntimes', arg_type='input_list', data_type=argument.Form(
    'jobRuntime', required_arguments= [
        Argument('runtimeType', choices=['DOCKER', 'SINGULARITY'])
    ],
    optional_arguments = [
        Argument('version')
    ]
    )),
    Argument('jobWorkingDir', default_value=r"HOST_EVAL($WORK2)", size_limit=(0, 4096), description='Where on this hpc system are jobs run?'),
])

USE_PROXY = argument.Form('useProxy', flattening_type='FLATTEN', required_arguments = [
    Argument('proxyHost', size_limit=(0, 256)),
    Argument('proxyPort', data_type='int'),
])

CAN_RUN_BATCH = argument.Form('canRunBatch', flattening_type='FLATTEN', required_arguments = [
    Argument('batchScheduler', choices=['SLURM', "CONDOR", "PBS", "SGE", "UGE", "TORQUE"], default_value='SLURM'),
    Argument('batchSchedulerProfile'),
    Argument('batchLogicalQueues', arg_type='input_list', data_type=argument.Form(
        'batchLogicalQueue', required_arguments = [
            Argument('name', size_limit=(1, 128)),
            Argument('hpcQueueName', size_limit=(1, 128)),
            Argument('maxJobs', data_type='int', default_value=50),
            Argument('maxJobsPerUser', data_type='int', default_value=10),
            Argument('minNodeCount', data_type='int', default_value=1),
            Argument('maxNodeCount', data_type='int', default_value=16),
            Argument('minCoresPerNode', data_type='int', default_value=1),
            Argument('maxCoresPerNode', data_type='int', default_value=16),
            Argument('minMemoryMB', data_type='int', default_value=1),
            Argument('maxMemoryMB', data_type='int', default_value=16384),
            Argument('minMinutes', data_type='int', default_value=1),
            Argument('maxMinutes', data_type='int', default_value=60)
        ]
    ))],
    optional_arguments = [
        Argument('batchDefaultLogicalQueue', size_limit=(1, 128)),
    ])

MOUNT_DATA_TRANSFER_NODE = argument.Form('mountDataTransferNode', flattening_type='RETRIEVE', required_arguments=[
    Argument('dtnSystemId', size_limit=(0, 80)),
    Argument('dtnMountPoint'),
    Argument('dtnMountSourcePath')
])

JOB_ENVIRONMENT_VARIABLES = Argument('jobEnvVariables', arg_type='input_list', data_type=argument.Form(
        'jobEnvironmentVariable', required_arguments = [
            Argument('key'),
            Argument('value', default_value=''),
        ],
        optional_arguments = [
            Argument('description', size_limit=(0, 2048), arg_type='str_input')
        ]
    ))


JOB_CAPABILITIES = Argument('jobCapabilities', arg_type='input_list', data_type=argument.Form(
            'jobCapability', required_arguments=[
                Argument('name', size_limit=(1, 128)),
                Argument('value'),
                Argument('category', choices=['SCHEUDLER', 
                                            'OS', 'HARDWARE', 
                                            'SOFTWARE', 'JOB', 
                                            'CONTAINER', 'MISC', 
                                            'CUSTOM']),
                Argument('datatype', choices=['STRING', 'INTEGER', 
                                            'BOOLEAN', 'NUMBER', 
                                            'TIMESTAMP']),
                Argument('precedence', data_type='int'),
        ]
    ))

JOB_CONSTRAINTS = argument.Form('jobConstraints', flattening_type='RETRIEVE', depends_on=['canExec'], required_arguments=[
        Argument('jobMaxJobs', data_type='int'),
        Argument('jobMaxJobsPerUser', data_type='int') # get rid of this form
    ])


import re
import inspect
import abc
import ast
import json
from typing import Type
import typing
from abc import abstractmethod, ABC
import os
from pprint import pprint, pformat
from datetime import datetime

# I finally understand. Imported at the top level by serverRun, so it can only see packages from that vantage point
from utilities import exceptions
from commands.arguments.argument import Argument, ALLOWED_ARG_TYPES
from socketopts import schemas
from commands import dataFormatters


__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(f"{__file__}")))
saved_command_root_dir = os.path.join(__location__, r'..\user_files')


def get_kwargs(function):
    sig = inspect.signature(function)
    keyword_args = [param.name for param in sig.parameters.values() if param.default != inspect.Parameter.empty]
    return keyword_args

def get_args(function):
    sig = inspect.signature(function)
    positional_args = [param.name for param in sig.parameters.values() if param.default == inspect.Parameter.empty]
    return positional_args


class HelpStringRetriever:
    def help_string_retriever(self):
        try:
            docstring_components = self.__class__.__doc__
            docstring_components = docstring_components.split("@")
            for docstring_component in docstring_components:
                if re.match(r'^[^:]+', docstring_component).group(0) == "help":
                    return docstring_component.split("help: ")[1]
            raise Exception()
        except Exception as e:
            raise exceptions.HelpDoesNotExist(self.__class__.__name__)
    

class CommandMetaClass(abc.ABCMeta):
    def __new__(cls, name, bases, attrs):
        instance = super().__new__(cls, name, bases, attrs)
        if name not in ('BaseQuery', 'BaseCommand'):
            instance.__check_run(name, attrs)
            instance.__check_command_args(name, attrs)
            instance.__check_command_opt(name, attrs)
        return instance
    
    def __check_run(self, name, attrs):
        if 'run' not in list(attrs.keys()):
            raise AttributeError(f"The command {name} requires a 'run()' method")
        elif not inspect.iscoroutinefunction(attrs['run']):
            raise AttributeError(f"The run method of the command {name} must be a coroutine")
    
    def __check_command_args(self, name, attrs):
        run_params = get_args(attrs['run'])
        args = []
        if 'required_arguments' in list(attrs.keys()): 
            args += attrs['required_arguments']
        if 'optional_arguments' in list(attrs.keys()): 
            args += attrs['optional_arguments']
        if args:
            for argument in args:
                if not issubclass(argument.__class__, Argument) and not isinstance(argument, Argument):
                    raise AttributeError(f"The argument {argument} of the command '{name}' must be of type 'Argument'. Currently {type(argument)}")
                if 'optional_arguments' in list(attrs.keys()) and argument in attrs['optional_arguments'] and argument.positional:
                    raise AttributeError(f"The optional argument {argument.argument} of the command {name} cannot be positional. All positional arguments must be required")
        if 'kwargs' not in run_params or 'args' not in run_params:
            raise AttributeError(f"The run() method of the {name} class must have a **kwargs and *args as parameters")
        
    def __check_command_opt(self, name, attrs):
        if 'command_opt' in list(attrs.keys()) and type((attrs['command_opt'])) != list:
            raise TypeError(f"The command opt attribute of the command {name} must be a list!")
        

class UpdatableFormRetriever(abc.ABC):
    @abc.abstractmethod
    def __call__(self, tapis_instance, **kwargs):
        pass


class BaseCommand(ABC, HelpStringRetriever, metaclass=CommandMetaClass):
    return_fields: list = []
    command_opt: list = None
    supports_config_file: bool = False
    required_arguments: list[Argument] | dict = list()
    optional_arguments: list[Argument] | dict = list()
    default_arguments = [Argument('connection', arg_type='silent'),
                            Argument('verbose', action='store_true', arg_type='silent'),
                            Argument('help', action='store_true', arg_type='silent'),
                            Argument('positionals', arg_type='silent')]
    updateable_form_retriever: UpdatableFormRetriever = None
    def __init__(self):
        self.t = None
        self.username = None
        self.password = None
        self.server = None
        self.arguments = dict()
        self.return_formatter: dataFormatters.BaseDataFormatter = dataFormatters.BaseDataFormatter(self.return_fields)
        self.required_arguments += self.default_arguments
        if self.supports_config_file:
            self.optional_arguments.append(Argument('file'))
        if self.required_arguments:
            for argument in self.required_arguments:
                argument.is_required(True)
        if self.optional_arguments:
            for argument in self.optional_arguments:
                argument.is_required(False)
        if isinstance(self.required_arguments, list):
            self.required_arguments = {argument.argument:argument for argument in self.required_arguments}
            self.arguments.update(**self.required_arguments)
        if isinstance(self.optional_arguments, list): 
            self.optional_arguments = {argument.argument:argument for argument in self.optional_arguments} 
            self.arguments.update(**self.optional_arguments)
        self.positional_arguments = [arg_name for arg_name, arg in self.required_arguments.items() if arg.positional]
        self.form_arguments = [arg_name for arg_name, arg in self.arguments.items() if arg.arg_type not in  ('standard', 'silent')]
        self.help: dict[dict[str, list[dict[str, str]]]] = dict()

        self.command_execution_sequence = []
        if self.positional_arguments:
            self.command_execution_sequence.append(self.check_for_positionals)
        if self.command_opt:
            self.command_execution_sequence.append(self.handle_arg_opts)
        self.command_execution_sequence.append(self.verify_argument_rules_followed)
        if self.form_arguments:
            self.command_execution_sequence.append(self.handle_form_input)
        self.command_execution_sequence.append(self.filter_kwargs)

    async def verify_argument_rules_followed(self, kwargs):
        """
        verify that the rules for each argument are followed, this is the argument validation
        """
        for name, value in self.arguments.items():
            kwargs[name] = self.arguments[name].verify_rules_followed(kwargs[name])
            if kwargs[name] and value.depends_on:
                for dependency in value.depends_on:
                    if not kwargs[dependency]:
                        raise Exception(f"The argument {name} requires the arguments {value.depends_on}")
            if kwargs[name] and value.mutually_exclusive_with and kwargs[value.mutually_exclusive_with]:
                raise Exception(f'The argument {name} is mutually exclusive with {value.mutually_exclusive_with}. Only one can be specified!')
        return kwargs

    async def filter_kwargs(self, kwargs):
        """
        filters out kwargs that have None value, Tapis breaks if I dont do this, especially with form arguments
        """
        filtered_kwargs = dict()
        for arg_name, value in kwargs.items():
            if value != None:
                filtered_kwargs[arg_name] = value
        #kwargs['connection'].logger.info(f"Running the command {self.__class__.__name__} with the kwargs {pformat(filtered_kwargs)}")
        return filtered_kwargs

    async def handle_config_file(self, kwargs):
        """
        when there is a config file submitted, that config file overrides all the other received kwargs
        """
        if kwargs['file']:
            with open(kwargs['file'], 'r') as f:
                kwargs = json.load(f)
            if 'error' in kwargs:
                kwargs.pop('error')
        return kwargs
    
    async def handle_form_input(self, kwargs):
        """
        handles forms when the user selects to fill them out
        """
        existing_values = dict() # If it aint broke, dont fix it. This code sucks but I cant meet the deadline if I try fixing this abomination
        if self.updateable_form_retriever:
            existing_values = self.return_formatter.obj_to_dict(self.updateable_form_retriever(self.t, **kwargs))
            to_pop = []
            reformatted_default_values = dict()
            for argument in self.form_arguments:
                argument = self.arguments[argument]
                if argument.arg_type == 'form' and argument.flattening_type in ('FLATTEN', 'RETRIEVE') and set(argument.arguments_list.keys()).issubset(set(existing_values.keys())):
                    reformatted_default_values[argument.argument] = {sub_arg_name:sub_arg_value for sub_arg_name, sub_arg_value in existing_values.items() if sub_arg_name in argument.arguments_list}
                    to_pop += list(argument.arguments_list.keys()) + [argument.argument]
                elif argument.arg_type == 'form' and argument.part_of and argument.argument in existing_values[argument.part_of]:
                    reformatted_default_values[argument.argument] = existing_values[argument.part_of][argument.argument]
                    if argument.part_of not in to_pop:
                        to_pop.append(argument.part_of)
            for argument, value in existing_values.items():
                if argument not in to_pop:
                    reformatted_default_values[argument] = value
            if reformatted_default_values:
                existing_values = reformatted_default_values

        for arg_name in self.form_arguments: 
            argument = self.arguments[arg_name]
            if kwargs[arg_name] or argument.required:
                pprint(f"NAME: {arg_name}\nVALUE: {kwargs[arg_name]}\n REQUIRED: {self.arguments[arg_name].required}")
                request = schemas.FormRequest(request_content={arg_name:argument}, existing_data=existing_values)
                await kwargs['connection'].send(request)
                response: schemas.FormResponse = await kwargs['connection'].receive()
                response_value = response.request_content
                if response_value == None:
                    raise ValueError(f"No value was received for the argument {arg_name}")
                response_value = argument.verify_rules_followed(response_value)
                if argument.arg_type == 'form' and argument.flattening_type in ('FLATTEN', 'RETRIEVE'):
                    kwargs.pop(arg_name)
                if argument.part_of:
                    if argument.part_of not in kwargs:
                        kwargs[argument.part_of] = dict()
                    kwargs[argument.part_of][argument.argument] = response_value
                elif argument.arg_type != 'form':
                    kwargs.update(**{arg_name:response_value})
                else:
                    kwargs.update(**response_value)
        return kwargs

    async def handle_arg_opts(self, kwargs):
        """
        arg opts handle special operations, like supporting relative file paths, and system entry
        """
        for opt in self.command_opt:
            kwargs = opt(kwargs)
        return kwargs
    
    async def check_for_positionals(self, kwargs):
        """
        some commands contain positional arguments. This function handles that circumstance
        """
        for arg_name, value in zip(self.positional_arguments, kwargs['positionals']):
            kwargs[arg_name] = value
        return kwargs

    def set_t_and_creds(self, t, username, password, server):
        """
        whenever the tenant or user changes, the new tapis object with the new credentials must be passed to each command to ensure they are operating on the currect user
        """
        self.t = t
        self.username = username
        self.password = password
        self.server = server

    def update_args_with_truncated(self, truncated_args_dict):
        """
        when the command group finishes processing all the truncated arguments, they get passed back here to be processed and assigned
        """
        try:
            for key in self.arguments:
                self.arguments[key].truncated_arg = f"-{truncated_args_dict[key]}"
            self.help['required'] = self.__argument_help_compiler(self.required_arguments)
            self.help['optional'] = self.__argument_help_compiler(self.optional_arguments)
        except Exception as e:
            raise e
        
    def __argument_help_compiler(self, arg_dict: dict[str, Argument]):
        """
        compile the help menus based on stored command metadata
        """
        verbose_dict = dict()
        standard_str = str()
        for name, argument in arg_dict.items():
            if argument.arg_type != 'silent' and not (argument.arg_type != 'standard' and argument.required):
                verbose_dict[name] = argument.help_message()
                standard_str += argument.str()
        return {'verbose':verbose_dict, 'standard':standard_str}

    def __get_help(self, verbose):
        """
        return the compiled help menus
        """
        help_dict = {"Command":self.__class__.__name__,
                    "Description":self.help_string_retriever(),
                    "Support Config File":self.supports_config_file}
        if not verbose:
            help_dict.update(**{
                "Syntax":f"{self.__class__.__name__} {self.help['required']['standard']}\n(Optional Arguments) {self.help['optional']['standard']}"})
        else:
            help_dict.update(**{
                "Syntax":f"{self.__class__.__name__}",
                "Required Arguments":self.help['required']['verbose'],
                "Optional Arguments":self.help['optional']['verbose']})
        return help_dict
    
    def brief_help(self):
        """
        return non verbose help
        """
        return {"Command":self.__class__.__name__,
                "Description":self.help_string_retriever()}
    
    @abstractmethod
    async def run(self, *args, **kwargs):
        """
        contains the actual command information at command definition
        """
        pass

    async def __call__(self, **kwargs):
        """
        runs all command meta-operations
        """
        kwargs['connection'].logger.info(f"Attempting to execute the command {self.__class__.__name__}")
        if self.supports_config_file and kwargs['file']: # if there is a config file input, skip all the other steps and take kwargs from that file
            new_kwargs = await self.handle_config_file(kwargs)
            for arg_name, arg in kwargs.items():
                if arg_name in ('connection', 'positionals', 'verbose', 'help'):
                    new_kwargs[arg_name] = arg
            kwargs = new_kwargs
        else:
            if kwargs['help']:
                return self.__get_help(verbose=kwargs['verbose'])
            for handler in self.command_execution_sequence:
                kwargs = await handler(kwargs)
        try:
            return_value = await self.run(**kwargs)
        except (exceptions.Shutdown, exceptions.Exit) as e:
            raise e
        except Exception as e: # this whole part writes command input to a file in the event a form command fails
            if self.supports_config_file:
                current_date = datetime.now()
                formatted_date = current_date.strftime("%m-%d-%y")
                main_save_path = f"{saved_command_root_dir}\\{formatted_date}"
                try:
                    os.listdir(main_save_path)
                except:
                    os.makedirs(main_save_path)
                number = len(os.listdir(main_save_path))
                file_save_path = f"{main_save_path}\\{self.__class__.__name__}_{number}.json"
                with open(file_save_path, 'w') as f:
                    kwargs.pop('connection')
                    kwargs.pop('positionals')
                    kwargs.pop('verbose')
                    kwargs.pop('help')
                    kwargs['error'] = str(e)
                    json.dump(kwargs, f, indent=4)
                    raise Exception(e, f"Argument input failure, command data written to file {file_save_path}")
            raise e
        return_value = self.return_formatter(return_value, kwargs['verbose'])
        return return_value
    

class BaseQuery(BaseCommand):
    def get_pod_credentials(self, id):
        uname, pword = self.t.pods.get_pod_credentials(pod_id=id).user_username, self.t.pods.get_pod_credentials(pod_id=id).user_password
        return uname, pword
    
    @abstractmethod
    async def run(self, *args, **kwargs):
        pass
    

class CommandMapMetaClass(type):
    def __new__(cls, name, bases, attrs):
        instance = super().__new__(cls, name, bases, attrs)
        if name not in ('CommandMapMetaClass', 'BaseCommandMap'):
            instance.__check_commands_are_proper_type(name, attrs)
            instance.__check_command_name(name, attrs)
            instance.__special_command_opts(name, attrs)
        return instance
    
    def __check_commands_are_proper_type(self, name, attrs):
        commands = attrs['command_map']
        if not commands:
            raise AttributeError(f"The command group {name} has no commands!")
            
    def __check_command_name(self, name, attrs):
        commands = attrs['command_map']
        for command_name, command in commands.items():
            if command_name != command.__class__.__name__:
                raise AttributeError(f"The command {command_name} in the command map {name} is a different name from its corresponding command class, {command.__class__.__name__}")

    def __special_command_opts(self, name, attrs):
        if 'command_opt' in list(attrs.keys()):
            if type(attrs['command_opt']) != list:
                raise TypeError(f"The command_opt attribute for the command group {name} must be a list!")
            for command_name, command in attrs['command_map'].items():
                if not command.command_opt:
                    command.command_opt = attrs['command_opt']
                    continue
                command.command_opt += attrs['command_opt']


class CommandContainer:
    aggregate_command_map: dict[str, Type[BaseCommand]] = dict()
    arguments : dict[str, Argument]= {
        'help':Argument('help', action='store_true'),
        'verbose':Argument('verbose', action='store_true'),
        'file':Argument('file')
    }


class BaseCommandMap(CommandContainer, HelpStringRetriever, metaclass=CommandMapMetaClass):
    command_map: dict[str, Type[BaseCommand]] = None
    command_opt = None
    def __init__(self):
        self.brief_help = self.__brief_help_gen()
        self.verbose_help = self.__help_gen()
        for name, command in self.command_map.items():
            self.aggregate_command_map.update({name:command})
            if command.required_arguments or command.optional_arguments:
                self.__contribute_to_arguments(command)

    def __contribute_to_arguments(self, command):
        arg_dict = {**command.required_arguments, **command.optional_arguments}
        for name, arg in arg_dict.items():
            if name not in self.arguments:
                self.arguments[name] = arg
            elif self.arguments[name].check_for_copy_data() != arg.check_for_copy_data():
                raise ValueError(f"Found two instances of the argument {arg.argument} that had different attributes.\n\n{command.__class__.__name__}: {arg.check_for_copy_data()}\n\nvs\n\n{self.arguments[name].check_for_copy_data()}")

    def __help_gen(self):
        verbose_help = list()
        for command in self.command_map.values():
            verbose_help.append(command.brief_help())
        return verbose_help
    
    def __brief_help_gen(self):
        brief_help = {
            'group':self.__class__.__name__,
            'description':self.help_string_retriever(),
            'instruction':f"enter {self.__class__.__name__} to retrieve list of commands"
        }
        return brief_help
    
    def __call__(self):
        return self.verbose_help



if __name__ != "__main__":
    from . import baseCommand
    from .arguments import argument
    from . import decorators
    from . import commandOpts
    Argument = argument.Argument


class ls(baseCommand.BaseCommand):
    """
    @help: list the files on a system 
    """
    command_opt = [commandOpts.CHECK_PWD(('file_path',)), commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('file_path', positional=True),
        Argument('systemId', size_limit=(1, 80), positional=True)
    ]
    async def run(self, *args, **kwargs) -> str: # lists files available on a tapis account
        file_list = self.t.files.listFiles(systemId=kwargs['systemId'], path=kwargs['file_path'])
        file_list_truncated = [f"{file.type} - {file.nativePermissions} ---- {file.name}" for file in file_list]
        return file_list_truncated
    

class cd(baseCommand.BaseCommand):
    """
    @help: change the directory
    """
    command_opt = [commandOpts.CHECK_PWD(('file_path',)), commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('file_path', positional=True),
        Argument('systemId', size_limit=(1, 80), positional=True)
    ]
    async def run(self, *args, **kwargs):
        self.t.files.listFiles(systemId=kwargs['systemId'], path=kwargs['file_path'])
        kwargs['connection'].pwd = kwargs['file_path']
        return kwargs['connection'].pwd
    

class pwd(baseCommand.BaseCommand):
    """
    @help: get the current directory
    """
    async def run(self, *args, **kwargs):
        return kwargs['connection'].pwd

class showme(baseCommand.BaseCommand):
    """
    @help: display file metadata
    """
    command_opt = [commandOpts.CHECK_PWD(('file_path',)), commandOpts.CHECK_EXPLICIT_ID('systemId')]
    return_fields = ['absolutePath', 'uid', 'size', 'perms']
    required_arguments = [
        Argument('file_path', positional=True),
        Argument('systemId', size_limit=(1, 80), positional=True)
    ]
    async def run(self, *args, **kwargs):
        return_data = self.t.files.getStatInfo(systemId=kwargs['systemId'], path=kwargs['file_path'])
        return return_data
    

class cat(baseCommand.BaseCommand):
    """
    @help: display small files directly to the terminal
    """
    command_opt = [commandOpts.CHECK_PWD(('file_path',)), commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('file_path', positional=True),
        Argument('systemId', size_limit=(1, 80), positional=True)
    ]
    async def run(self, *args, **kwargs):
        size = self.t.files.getStatInfo(systemId=kwargs['systemId'], path=kwargs['file_path']).size
        if size <= 4096:
            file_info = self.t.files.getContents(systemId=kwargs['systemId'],
                                path=kwargs['file_path'])
        else:
            file_info = "file too large to print"
        return file_info
    

class mkdir(baseCommand.BaseCommand):
    """
    @help: create a new directory at the selected path
    """
    command_opt = [commandOpts.CHECK_PWD(('file_path',)), commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('file_path', positional=True),
        Argument('systemId', size_limit=(1, 80), positional=True)
    ]
    async def run(self, *args, **kwargs):
        self.t.files.mkdir(systemId=kwargs['systemId'], path=kwargs['file_path'])
        return f"Successfully created file at {kwargs['file_path']}"
    

class mv(baseCommand.BaseCommand):
    """
    @help: move a file from a source directory to a destination directory within a system's file structure
    """
    command_opt = [commandOpts.CHECK_PWD(('source_file', 'destination_file')), commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('source_file', positional=True),
        Argument('systemId', size_limit=(1, 80), positional=True),
        Argument('desination_file')
    ]
    async def run(self, destination_file: str, *args, **kwargs):
        self.t.files.moveCopy(systemId=kwargs['systemId'], path=kwargs['source_file'], operation="MOVE", newPath=destination_file)
        return f"File moved successfully to {destination_file}"
    

class cp(baseCommand.BaseCommand):
    """
    @help: copy a file from a source directory to another directory
    """
    command_opt = [commandOpts.CHECK_PWD(('source_file', 'destination_file')), commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('source_file', positional=True),
        Argument('systemId', size_limit=(1, 80), positional=True),
        Argument('destination_file')
    ]
    async def run(self, *args, **kwargs):
        self.t.files.moveCopy(systemId=kwargs['systemId'], path=kwargs['source_file'], operation="COPY", newPath=kwargs['destination_file'])
        return f"File copied successfully to {kwargs['destination_file']}"
    

class rm(baseCommand.BaseCommand):
    """
    @help: delete a selected file
    """
    command_opt = [commandOpts.CHECK_PWD(('file_path',)), commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('file_path', positional=True),
        Argument('systemId', size_limit=(1, 80), positional=True)
    ]
    async def run(self, *args, **kwargs):
        self.t.delete(systemId=kwargs['systemId'], path=kwargs['file_path'])
        return f"file {kwargs['file_path']} successfully deleted"
    

class get_recent_transfers(baseCommand.BaseCommand):
    """
    @help: return a list of recent file transfers
    """
    return_fields = ['id', 'username', 'status', 'estimatedTotalBytes']
    async def run(self, *args, **kwargs):
        recent_transfers = self.t.files.getRecentTransferTasks()
        return recent_transfers
    

class create_transfer_task(baseCommand.BaseCommand):
    """
    @help: create a task to transfer files between systems. The two system types must be the same
    """
    required_arguments = [
        Argument('source_file_path', positional=True),
        Argument('source_system_id', positional=True),
        Argument('destination_system_id'),
        Argument('destination_file_path')
    ]
    async def run(self, *args, **kwargs):
        pass


class grant_permissions(baseCommand.BaseCommand):
    """
    @help: grant permission to a file path for a user
    """
    return_fields = ['username', 'permission', 'path']
    command_opt = [commandOpts.CHECK_PWD(('path',)), commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('path', positional=True),
        Argument('systemId', positional=True),
        Argument('username'),
        Argument('permission', choices=['READ', 'MODIFY'])
    ]
    async def run(self, *args, **kwargs):
        return self.t.files.grantPermissions(**kwargs)
    

class get_permissions(grant_permissions):
    """
    @help: see what permissions users have on a file or path
    """
    required_arguments = [
        Argument('path', positional=True),
        Argument('systemId', positional=True)
    ]
    async def run(self, *args, **kwargs):
        return self.t.files.getPermissions(**kwargs)
    

class delete_permissions(baseCommand.BaseCommand):
    """
    @help: delete permissions to a filepath
    """
    command_opt = [commandOpts.CHECK_PWD(('path',)), commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('path', positional=True),
        Argument('systemId', positional=True),
        Argument('username')
    ]
    async def run(self, *args, **kwargs):
        return self.t.files.deletePermissions(**kwargs)


class create_postit(baseCommand.BaseCommand):
    """
    @help: create a postit to easily share file with other users
    """
    return_fields = ['postitId', 'systemId', 'path', 'redeemUrl']
    command_opt = [commandOpts.CHECK_PWD(('file_path',)), commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('file_path', positional=True),
        Argument('systemId', size_limit=(1, 80), positional=True)
    ]
    optional_arguments = [
        Argument('allowed_uses', data_type='int'),
        Argument('expiration_time', data_type='int')
    ]
    async def run(self, *args, **kwargs):
        self.t.files.createPostIt(systemId=kwargs['systemId'], path=kwargs['file_path'], allowedUses=kwargs['allowed_uses'], validSeconds=kwargs['expiration_time'])
        return f"created a postit for the file {kwargs['file_path']}"
    

class list_postits(baseCommand.BaseCommand):
    """
    @help: list all postits
    """
    return_fields = ['postitId', 'systemId', 'path', 'redeemUrl']
    optional_arguments = [
        Argument('owned_or_all', choices=['ALL', 'OWNED'])
    ]
    async def run(self, *args, **kwargs):
        postit_list = str(self.t.files.listPostIts(listType=kwargs['owned_or_all']))
        return postit_list
    

class get_postit(baseCommand.BaseCommand):
    """
    @help: get a specific postit information based on postit id
    """
    return_fields = ['postitId', 'systemId', 'path', 'redeemUrl']
    required_arguments = [
        Argument('postitId', positional=True),
    ]
    async def run(self, *args, **kwargs):
        postit = str(self.t.files.getPostIt(postitId=kwargs['postitId']))
        return postit
    

class delete_postit(baseCommand.BaseCommand):
    """
    @help: get a specific postit information based on postit kwargs['systemId']
    """
    required_arguments = [
        Argument('postitId', positional=True),
    ]
    async def run(self, *args, **kwargs):
        postit = str(self.t.files.deletePostIt(postitId=kwargs['postitId']))
        return f"Successfully deleted postit {kwargs['postitId']}"
    

class redeem_postit(baseCommand.BaseCommand):
    """
    @help: redeem a postit and download posted file. Downloads to browser
    """
    required_arguments = [
        Argument('postitId', positional=True),
    ]
    async def run(self, *args, **kwargs):
        self.t.files.redeemPostIt(postitId=kwargs['postitId'])
        return f"Downloading postit {kwargs['postitId']}, check browser"

    
class upload(baseCommand.BaseCommand):
    """
    @help: upload a file to the system 
    the source and destination files must both be in the file argument, respectively, separated by a comma
    @todo: make it so that this doesnt need to take both source and destination files, but have it so it retrieves the current file location on the tapis system
    and sets that file location to be the upload point. Do the same for downloads but in reverse
    """
    command_opt = [commandOpts.CHECK_PWD(('destination_file',)), commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('source_file', positional=True),
        Argument('systemId', size_limit=(1, 80), positional=True),
        Argument('destination_file'),
    ]
    async def run(self, *args, **kwargs) -> str: # upload a file from local to remote using tapis. Takes source and destination paths
        if not kwargs['destination_file']:
            kwargs['destination_file'] = kwargs['connection'].pwd
        self.t.upload(system_id=kwargs['systemId'],
                source_file_path=kwargs['source_file'],
                dest_file_path=kwargs['destination_file'])
        return f'successfully uploaded {kwargs["source_file"]} to {kwargs["destination_file"]}'


class download(baseCommand.BaseCommand):
    """
    @help: download a file from the system
    the source and destination files must both be in the file argument, respectively, separated by a comma
    """
    command_opt = [commandOpts.CHECK_PWD(('source_file',)), commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('source_file', positional=True),
        Argument('systemId', size_limit=(1, 80), positional=True),
        Argument('destination_file'),
        Argument('connection', arg_type='silent')
    ]
    async def run(self, *args, **kwargs) -> str: # download a remote file using tapis, operates basically the same as upload
        if not kwargs["source_file"]:
            kwargs["source_file"] = kwargs['connection'].pwd
        file_info = self.t.files.getContents(systemId=kwargs['systemId'],
                            path=kwargs["source_file"])
        file_info = file_info.decode('utf-8')
        with open(kwargs['destination_file'], 'w') as f:
            f.write(file_info)
        return f'successfully downloaded {kwargs["source_file"]} to {kwargs["destination_file"]}'
    




import time
import socket
import asyncio
import traceback
import typing
import os
import json

from tapipy.tapis import Tapis

from commands import commandMap, decorators
from utilities import logger, exceptions
from socketopts import schemas, socketOpts
from server import auth


class ServerConnection(socketOpts.ServerSocketOpts):
    """
    connection object to wrap around async reader and writer to make work easier
    """
    def __init__(self, name, reader, writer, connection_list, debug=False):
        super().__init__(name, debug=debug)
        self.name = name
        self.connection_list: list = connection_list
        self.reader: asyncio.StreamReader = reader
        self.writer: asyncio.StreamWriter = writer
        self.task: asyncio.Task = None
        self.status = "CLOSED"
        self.shutdown_message = schemas.ResponseData(message={'message':'Shutdown initiated, closing'}, exit_status=1)
        self.logger.info('Connection successfully initiated, beginning handshake')

    def set_status_device_authenticating(self):
        self.status = 'DEVICE_CODE_AUTH'

    def set_status_closed(self):
        self.status = 'CLOSED'

    def set_status_exiting(self):
        self.status = 'EXITING'

    def set_task(self, task: asyncio.Task):
        self.task = task
        self.status = 'OPEN'
        self.logger.info('Asyncio task for socket operations was received, and is now active')

    async def close(self, connection_list_lock: asyncio.Lock):
        self.logger.info('ATTEMPTING TO CLOSE')
        if self.status in ('OPEN', 'EXITING'):
            result = self.task.cancel()
            self.logger.info("successfully cancelled task")
            await self.send(self.shutdown_message)
            self.writer.close()
            await self.writer.wait_closed()
            self.status = 'CLOSED'
        elif self.status == 'DEVICE_CODE_AUTH': # only in the auth file
            await self.send(schemas.AuthRequest(error='cancelled authentication during device_code grant', auth_request_type='device_code'))
            self.writer.close()
            await self.writer.wait_closed()
        else:
            self.writer.close()
            await self.writer.wait_closed()
        async with connection_list_lock:
            try:
                self.connection_list.remove(self)
            except Exception as e:
                self.logger.info("Connection was not in the command list, closure ocurred during connection setup")
        print(self.connection_list)
        self.logger.info('SUCCESSFULLY CLOSED')


class Server(commandMap.AggregateCommandMap, logger.ServerLogger, decorators.DecoratorSetup, auth.ServerSideAuth):
    """
    Receives commands from the client and executes Tapis operations
    """
    SESSION_TIME = 5000
    debug=False
    def __init__(self, IP: str, PORT: int):
        super().__init__()
        self.initial = True
        self.running = True

        self.t = None
        self.url = None
        self.access_token = None
        self.username = None
        self.password = None
        self.auth_type = None

        self.__name__ = "Server"
        self.initialize_logger(self.__name__)
        # setting up socket server
        self.ip, self.port = IP, PORT
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setblocking(False)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind((self.ip, self.port))
        self.sock.listen(1)

        self.end_time = time.time() + self.SESSION_TIME # start the countdown on the timeout

        self.connections_list: list[ServerConnection] = []

        self.server = None
        self.num_connections = 0

        self.connection_list_lock = asyncio.Lock()
        self.logger.info('initialization complete')

    async def close_connections(self):
        print([connection.status for connection in self.connections_list])
        return await asyncio.gather(*[connection.close(self.connection_list_lock) for connection in self.connections_list])

    async def close(self):
        await self.close_connections()
        self.server.close()
        raise exceptions.Shutdown()

    async def check_timeout(self):
        while self.running:
            if time.time() >= self.end_time:
                self.logger.info("Timeout, shutting down")
                self.running = False
                return 'server timed out'
            await asyncio.sleep(3)
    
    async def check_shutdown(self):
        while self.running:
            await asyncio.sleep(3)
            for connection in self.connections_list:
                if connection.status == 'EXITING':
                    await connection.close(self.connection_list_lock)
        await self.close()
        self.logger.info("The server shutdown.")
        return None

    async def handshake(self, connection):
        self.logger.info("Handshake starting")
        if self.initial:  # if this is the first time in the session that the cli is connecting
            startup_data = schemas.StartupData(initial = self.initial)
            await connection.send(startup_data)
            await self.auth_startup(connection)
        else:
            startup_result = schemas.StartupData(initial = self.initial, username = self.username, url = self.url)
            await connection.send(startup_result)
        self.initial = False
        self.logger.info("Final connection data sent")

    async def accept(self, reader, writer):
        """
        accept connection request and initialize communication with the client
        """  
        self.num_connections += 1
        self.end_time = time.time() + self.SESSION_TIME
        connection = ServerConnection(f"CON-{self.num_connections}", reader=reader, writer=writer, connection_list=self.connections_list, debug=self.debug)
        ip, port = writer.transport.get_extra_info('socket').getsockname()
        try:
            await self.handshake(connection)
            connection.logger.info('Handshake complete')
        except exceptions.InvalidCredentialsReceived as e:
            connection.logger.warning("invalid credentials entered too many times. Cancelling request")
            await connection.close(self.connection_list_lock)
            return
        except exceptions.ClientSideError as e:
            connection.logger.warning(f"Encountered client side error during startup handshake. {e}")
            await connection.close(self.connection_list_lock)
            return
        except Exception as e:
            connection.logger.warning(e)
            await connection.close(self.connection_list_lock)
            return
        connection.logger.info('Sending command arguments now')
        await connection.send(schemas.ResponseData(request_content={name:argument.json() for name, argument in self.arguments.items()}))
        setup_response: schemas.ResponseData = await connection.receive()
        if not setup_response.request_content['setup_success']:
            connection.logger.warning(f"The setup of the connection upon sending argument information failed")
            return
        connection.logger.info(f"Argument setup succeeded on the client side")
        loop = asyncio.get_event_loop()
        task: asyncio.Task = loop.create_task(self.receive_and_execute(connection))
        connection.set_task(task)
        self.connections_list.append(connection)
        connection.logger.info('All connection setup complete, beginning normal operations')

    async def receive_and_execute(self, connection: ServerConnection):
        """
        receive and process commands
        """
        while self.running:
            try:
                message = await connection.receive()
                if not self.running:
                    raise exceptions.Shutdown
                kwargs = message.request_content
                result = await self.run_command(connection, kwargs)
                response = schemas.ResponseData(message={"message":result}, url=self.url, active_username=self.username)
                self.end_time = time.time() + self.SESSION_TIME 
                await connection.send(response)
            except exceptions.ClientSideError as e:
                self.logger.warning(e)
                continue
            except (exceptions.TimeoutError, exceptions.Shutdown) as e:
                connection.logger.warning(str(e))
                self.running = False
                return
            except exceptions.Exit as e:
                connection.logger.info("user exit initiated")
                connection.set_status_exiting()
                return
            except asyncio.CancelledError:
                return 'task was cancelled'
            except (exceptions.CommandNotFoundError, exceptions.NoConfirmationError, exceptions.InvalidCredentialsReceived, Exception) as e:
                if isinstance(e, OSError) and 'The specified network name is no longer available' in str(e):
                    print(e)
                    connection.logger.info("connection was lost, waiting to reconnect")
                    connection.set_status_closed()
                    return
                error_str = traceback.format_exc()
                error_response = schemas.ResponseData(error=str(e), url=self.url, active_username=self.username)
                await connection.send(error_response)
                connection.logger.warning(f"{error_str}")
    
    async def main(self):
        self.server = await asyncio.start_server(self.accept, sock=self.sock)
        try:
            async with self.server:
                result = await asyncio.gather(self.server.serve_forever(), self.check_timeout(), self.check_shutdown(), return_exceptions=True)#, self.check_timeout(), return_exceptions=True)
                self.logger.info(str(result))
        except (KeyboardInterrupt, exceptions.Shutdown):
            self.running = False


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    server = Server(socket.gethostbyname('127.0.0.1', 30000)) 
    try:
        asyncio.run(server.main())
    finally:
        loop.close()
